/*==========================================================================*\
 |  This file is part of setiherder (http://www.sackheads.org/setiherder/)  |
 |  which was written by Brian L. Naylor <brian+seti@sackheads.org> and is  |
 |  under the GNU General Public License.  You should have recieved a copy  |
 |  of the GPL with this source code.  If not, we'd like to know about it.  |
\*--------------------------------------------------------------------------*/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <time.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "shnet.h"
#include "shutil.h"
#include "shclients.h"
#include "shqueue.h"

short seq_list[SH_MAX_CLIENTS];

int shnet_init()
{
  /* set up data */
  memset((void *)&seq_list, 0, sizeof(short) * SH_MAX_CLIENTS);
  return SH_SUCCESS;
}

short shnet_next_seq(int idx)
{
  if(seq_list[idx] == 255)
    seq_list[idx] = 0;
  else
    seq_list[idx]++;

  return(seq_list[idx]);
}

int shnet_connect(sh_client_t *cli)
{
  struct hostent      *host;
  struct protoent     *prot;
  struct sockaddr_in   sin;
  int    s;
 
  /*
   *  sanity check - is this thing set to use the network?
   */
  if(cli->spawn_method != SH_SPAWN_NET)
  {
    dbg(DBG_ERROR, "shnet_connect(): client %d (%s) is not set up to "
                   "use setid\n",
                    cli->index, cli->hostname);
    return SH_ERR_NOTWERK;
  }
  
  memset((void *)&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;

  if((sin.sin_port = htons((u_short)cli->port)) == 0)
  {
    dbg(DBG_ERROR, "shnet_connect(): client %d (%s) has invalid port %d\n",
                   cli->index, cli->hostname, cli->port);
    return SH_ERR_NOTWERK;
  }

  host = gethostbyname(cli->hostname);
  if(host)
  {
    if(!host->h_addr_list[0])
    {
      dbg(DBG_ERROR, "Host %s does not have an IP address!\n", cli->hostname);
      return SH_ERR_NOTWERK;
    }
    sin.sin_family = host->h_addrtype;
    memcpy((void *)&sin.sin_addr, 
           (void *)host->h_addr_list[0], 
           sizeof(sin.sin_addr));
  }
  else if((sin.sin_addr.s_addr = inet_addr(cli->hostname)) == INADDR_NONE)
  {
    dbg(DBG_ERROR, "shnet_connect(): can't resolve '%s'\n",
                   cli->hostname);
    return SH_ERR_NOTWERK;
  }
  if((prot = getprotobyname("tcp")) == NULL)
  {
    dbg(DBG_ERROR, "shnet_connect(): woo, getprotobyname() failed.  huh?\n");
    return SH_ERR_NOTWERK;
  }
  
  if((s = socket(PF_INET, SOCK_STREAM, prot->p_proto)) < 0)
  {
    dbg(DBG_ERROR, "shnet_connect(): couldn't allocate socket\n");
    return SH_ERR_NOTWERK;
  }

  if(connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0)
  {
    dbg(DBG_ERROR, "shnet_connect(): can't connect to %s:%d, rc=%d (%s)\n",
                   cli->hostname, cli->port, errno, strerror(errno));
    return SH_ERR_NOTWERK;
  }

  cli->ptyfp = fdopen(s, "rw");
  cli->socket = s;

  if(cli->passwd[0] != '\0')
    shnet_send_passwd(cli);
  shnet_set_detach(cli);
  shnet_send_cmd(cli, SH_NET_CMD_VERSION);
  
  return SH_SUCCESS;
}

int shnet_close(sh_client_t *cli)
{
  if(cli->iotag != 0) 
    gdk_input_remove(cli->iotag); 
  cli->iotag = 0;
  cli->authed = 0;
  shutdown(cli->socket, 1);
  close(cli->socket);
  return SH_SUCCESS;
}

void shnet_idle()
{
  static int idx = 0;

}

int shnet_parse(sh_client_t *cli, char *line)
{
  char         *ptr;
  sh_q_item_t **orig=NULL;
  int           seq;
  
  printf("shnet_parse() got(%d): %s\n",
         cli->index, line);

  ptr = strchr(line, ':');
  if(ptr == NULL)
  {
    dbg(DBG_ERROR, "shnet_parse() got bogus input.\n");
    dbg(DBG_DUMP, "-> %s\n", line);
    return SH_ERR_GENERIC;
  }
  seq = atoi(++ptr);
  
  if(seq >= 0)
    sh_q_got_input(cli->index, seq, orig);
  
  if(orig != NULL)
  /* XXX can use 'orig' for parsing context! XXX */
    free(orig);

  if(strstr(line, SH_NET_ERR_AUTHREQ))
  {
    if(cli->authed)
    {
      dbg(DBG_ERROR, "Auth failed for client %d (%s).\n", 
          cli->index, cli->hostname);
      return SH_SUCCESS;
    }
    else if(cli->passwd[0] == '\0')
    {
      dbg(DBG_ERROR, "Client %d (%s) requires a password.\n",
          cli->index, cli->hostname);
      return SH_SUCCESS;
    } 
    else
    {
      cli->authed = 1;
      return(shnet_send_passwd(cli));
    }
  }
  else if(strstr(line, SH_NET_ERR_OK))
  {
    return SH_SUCCESS;
  }
  else if(strstr(line, SH_NET_ERR_PONG))
  {
    dbg(DBG_CUTE, "Pong from %s\n", cli->hostname);
  }
  else if(strstr(line, SH_NET_ERR_VERSION))
  {
    int major, minor;

    char *getnext(char delim) 
    {
      ptr = strchr(ptr, delim);
      if(ptr == NULL)
      {
        dbg(DBG_ERROR, "Client %d (%s) returned an invalid version string.\n",
            cli->index, cli->hostname);
      }
      return(++ptr);
    }

    ptr = line;
    major = atoi(line);
    minor = atoi(getnext('.'));
    if( (major < SH_VERSION_MAJOR) ||
        (minor < SH_VERSION_MINOR) )
    {
      dbg(DBG_WARN, 
          "Client %d (%s) is downlevel, from setiherder version %d.%d\n",
          cli->index, cli->hostname, major, minor);
      dbg(DBG_WARN, 
          "It may lack certain newer features supported by this client.\n");
    }
    major = atoi(getnext(' '));
    minor = atoi(getnext('.'));
    if( (major < SH_NET_VERSION_MAJOR) ||
        (minor < SH_NET_VERSION_MINOR) )
    {
      dbg(DBG_WARN,
          "Client %d (%s) is using a downlevel network protocol, version %d.%d\n",
          cli->index, cli->hostname, major, minor);
      dbg(DBG_WARN, 
          "It may not be able to communicate fully with this client.\n");
    }
  }
  else if(strstr(line, SH_NET_ERR_ACTIVE))
  {
    dbg(DBG_WARN, "Client %d (%s) is already active.\n",
        cli->index, cli->hostname);
  }
  else if(strstr(line, SH_NET_ERR_NACTIVE))
  {
    dbg(DBG_WARN, "Client %d (%s) is not active.\n",
        cli->index, cli->hostname);
  }
  else if(strstr(line, SH_NET_ERR_CONNECT))
  {
    dbg(DBG_ERROR, "Client %d (%s) already has a setiherder attached!\n",
        cli->index, cli->hostname);
    shnet_close(cli);
  }
  else if(strstr(line, SH_NET_ERR_CLIENT_START))
  {
    dbg(DBG_ERROR, "Client %d (%s) was unable to start seti@home.\n",
        cli->index, cli->hostname);
    shnet_close(cli);
    cli->pid   = SH_DEFAULT_DEAD_PID;
    cli->state = SH_STATE_STOPPED;
    cli->died  = 0; /* something is wrong here.. */
    shwin_status_change(cli->index);
  }
  else if(strstr(line, SH_NET_ERR_CLIENT_COM))
  {
    dbg(DBG_ERROR, "Client %d (%s) reports that seti@home is not responding.\n",
        cli->index, cli->hostname);
  }
  else if(strstr(line, SH_NET_ERR_HOST))
  {
    dbg(DBG_ERROR, "Client %d (%s) does not have this host in its access list.\n",
        cli->index, cli->hostname);
    shnet_close(cli);
  }
  else if(strstr(line, SH_NET_ERR_DIED))
  {
    dbg(DBG_WARN, "Client %d (%s) has terminated.\n",
        cli->index, cli->hostname);
    shnet_close(cli);
    cli->pid    = SH_DEFAULT_DEAD_PID;
    cli->state  = SH_STATE_STOPPED;
    cli->died   = 0;
    /* don't mark it 'dead' if the user stopped it. */
    if(cli->user_stopped == 0)
    {
      cli->died  = time(NULL);
      cli->state = SH_STATE_CRASHED;
    }
    shwin_status_change(cli->index);
  }
  else if(strstr(line, SH_NET_ERR_SHUTDOWN))
  {
    dbg(DBG_ERROR, "Client %d (%s)'s setid is exiting!\n",
        cli->index, cli->hostname);
    shnet_close(cli);
    cli->pid   = SH_DEFAULT_DEAD_PID;
    cli->state = SH_STATE_STOPPED;
    cli->died  = 0; /* no use restarting.. */
    shwin_status_change(cli->index);
  }
  else if(strstr(line, SH_NET_ERR_SYNTAX))
  {
    dbg(DBG_ERROR, "Client %d (%s) and I do not agree on syntax.  This is bad.\n",
        cli->index, cli->hostname);
  } 
  else if(strstr(line, SH_NET_ERR_GENERIC))
  {
    dbg(DBG_ERROR, "Client %d (%s) is confused about something, and is spouting"
        " generic errors.\n", cli->index, cli->hostname);
  }
  else if(ptr = (strstr(line, SH_NET_CMD_IDLE)))
  {
    int inc, i;
    ptr = strchr(line, ' ');
    inc = atoi(++ptr);
    dbg(DBG_INFO, "Client %d (%s) has completed %d work units since we last "
        "spoke to it.\n", cli->index, cli->hostname, inc);
    for(i=0; i<inc; i++)
      sh_client_increment_work(cli->index);
  }

  return SH_SUCCESS;
} 
 
int shnet_send_cmd(sh_client_t *cli, char *cmd)
{
  char        *tmp;
  sh_q_item_t *pkt;
  int          seq;
 
  if((tmp = (char *)malloc(SH_MAX_LINE + 1)) == NULL)
  {
    dbg(DBG_FATAL, "shnet_send_cmd() couldn't allocate %d bytes!\n",
        SH_MAX_LINE + 1);
    return SH_ERR_MEM;
  }
  if((pkt = (sh_q_item_t *)malloc(sizeof(sh_q_item_t))) == NULL)
  {
    dbg(DBG_FATAL, "shnet_send_cmd() couldn't allocate %d bytes!\n",
        sizeof(sh_q_item_t));
    return SH_ERR_MEM;
  }
  
  seq = shnet_next_seq(cli->index);
#ifdef HAVE_SNPRINTF  
  snprintf(tmp, 8, "%s%d:", SH_NET_PREFIX, seq);
#else  
  sprintf(tmp, "%s%d:", SH_NET_PREFIX, seq);
#endif  
  strncat(tmp, cmd, SH_MAX_LINE - strlen(tmp) - 1);
  strcat (tmp, "\n");

  pkt->sent     = 0;
  pkt->idle_cnt = 0;
  pkt->text     = tmp;
  pkt->next     = NULL;
  pkt->seq      = seq;

  return(sh_q_append(cli->index, pkt));
}

int shnet_raw_send(int idx, char *msg)
{
  sh_client_t *sc = sh_client_get(idx);
  int rc;

  printf("sending: %s", msg);

  if(sc == NULL)
  {
    dbg(DBG_ERROR, "shnet_raw_send() got invalid index number (%d)!\n", idx);
    return SH_ERR_GENERIC;
  }

  if(sc->iotag == 0)
  {
    dbg(DBG_ERROR, "Cannot send command to %s; not connected.\n",
        sc->hostname);
    return SH_ERR_NOTWERK;
  }
  
  if((rc = write(sc->socket, msg, strlen(msg))) < 0)
  {
    dbg(DBG_ERROR, "shnet_raw_send() Couldn't write to socket (%s)\n",
        strerror(rc));
    return SH_ERR_NOTWERK;
  }

  return SH_SUCCESS;
}


int shnet_set_detach(sh_client_t *cli)
{
  static char tmp[80];
  
  sprintf(tmp, "%s %d", SH_NET_CMD_DETACH, cli->detach);
  return(shnet_send_cmd(cli, tmp));
}

int shnet_send_passwd(sh_client_t *cli)
{
  char *tmp;
  int   rc;
 
  if(cli == NULL || cli->passwd == NULL) 
  {
    return SH_ERR_GENERIC;
  }
  
  if((tmp = (char *)malloc(strlen(cli->passwd) + strlen(SH_NET_CMD_AUTH) + 1)) 
      == NULL)
  {
    return SH_ERR_MEM;
  }
  
  sprintf(tmp, "%s %s", SH_NET_CMD_AUTH, cli->passwd);
  rc = shnet_send_cmd(cli, tmp);
  free(tmp);
  return rc;
}

int shnet_ping(sh_client_t *cli)
{
  int          rc;
  sh_q_item_t *newpkt;
  char        *msg;
  
  if((newpkt = (sh_q_item_t *)malloc(sizeof(sh_q_item_t))) == NULL)
  {
    dbg(DBG_ERROR, "shnet_ping() couldn't allocate %d bytes!\n",
        sizeof(sh_q_item_t));
    return SH_ERR_MEM;
  }
  
  if((msg = (char *)malloc( strlen(SH_NET_PREFIX)   + 4 /* ':' + 3 digits */ +
                            strlen(SH_NET_CMD_PING) + 2 /* \n\0           */
                          )) == NULL)
  {
    dbg(DBG_ERROR, "shnet_ping() couldn't allocate space for ping text!\n");
    return SH_ERR_MEM;
  }
  
  newpkt->seq = shnet_next_seq(cli->index);
  sprintf(msg, "%s%d:%s\n", SH_NET_PREFIX, newpkt->seq, SH_NET_CMD_PING);
  if((rc = shnet_raw_send(cli->index, msg)) == SH_SUCCESS)
  {
    newpkt->sent       = 1;
    newpkt->text       = msg;
    newpkt->next       = queues[cli->index];
    newpkt->idle_cnt   = 0;
    queues[cli->index] = newpkt;
  }
  else
  {
    free(newpkt);
    free(msg);
    return rc;
  }
  
  return SH_SUCCESS;
}

