/*==========================================================================*\
 |  This file is part of setiherder (http://www.sackheads.org/setiherder/)  |
 |  which was written by Brian L. Naylor <brian+seti@sackheads.org> and is  |
 |  under the GNU General Public License.  You should have recieved a copy  |
 |  of the GPL with this source code.  If not, we'd like to know about it.  |
\*--------------------------------------------------------------------------*/

#ifndef SH_QUEUE
#define SH_QUEUE

/*--------------------------------------------------------------------------*\
 | description of queue operation                                           |
 |                                                                          |
 | This client uses non-blocking sockets and can receive unsolicited        |
 | packets from its clients, but does need to be able to match up replies   |
 | to responses, etc.  So, we use a transmission queue to enforce a metered |
 | transmission rate, to match replies, and to handle timeout, pings,       |
 | retransmission, etc.                                                     |
 |                                                                          |
 | Each setiherder<->setid control packet has a sequence number.  Packets   |
 | originating from the gui range between 1 and 255, while those from the   |
 | daemon are negative.  Each setid client has its own queue that tracks:   |
 |   sequence number                                                        |
 |   how many times the idle loop has checked it w/o a response             |
 |   whether the command has actually been sent yet                         |
 |   text of command                                                        |
 |                                                                          |
 | When the gui wants to send a command, it simply tacks the command onto   |
 | the end of the queue, and the idle loop will actually do the network     |
 | transmission when there are no other outstanding packets to be sent      |
 | before it.  (Ping is considered 'OOB' - it will be prepended onto the    |
 | queue and processed immediately.)                                        |
 |                                                                          |
 | If the idle timer revisits a queue entry SH_Q_MAX_WAIT times and the     |
 | packet has not been answered, it will prepend a ping to the queue, send  |
 | it, and iterate once more.  If the ping is answered, then we know        |
 | something went wrong, and can try a retransmission of the original       |
 | packet (or issue an error...)  If the ping does not come back, then we   |
 | have a more serious error (severe network lag or setid hang) and should  |
 | indicate as much to the user.                                            |
 |                                                                          |
 | Responses from the client are not processed by the idle loop.  These are |
 | routed in directly when GTK+'s input monitoring system receives them.    |
 | shnet_parse() informs the queue of the receipt of the packet by calling  |
 | sh_q_got_input() if the sequence number is positive.  (If it's a daemon- |
 | initiated packet, then the queue doesn't have an entry for it anyway and |
 | doesn't need to know about it.)  That function steps through the queue   |
 | for the client in question and tries to match the sequence number with   |
 | a transmitted packet.  If it                                             |
 |   - finds it, removes the queue entry, transmits the next command if     |
 |     there is one, and returns SH_SUCCESS and the matching queue entry    |
 |     in '**orig'.  The caller must free() the queue item.                 |
 |   - doesn't find it, returns a serious error.                            |
\*--------------------------------------------------------------------------*/

#include "shtypes.h"

typedef struct sh_q_item {
  short   seq;             /* packet sequence number                       */
  short   idle_cnt;        /* # of times checked by idle loop w/o response */
  short   sent;            /* flag: has the packet has been transmitted?   */
  char   *text;            /* pointer to entire text of sent command       */
  struct  sh_q_item* next; /* pointer to next queue entry                  */
} sh_q_item_t;

/*
 * Default initial timer interval.  (How often does the queue get checked?)
 */
#define SH_Q_DFLT_INTERVAL   2000 /* milliseconds */

/*
 * How many times will the idle loop check an "unrequited" packet without
 * taking any sort of action?  If idle_cnt hits MAX_WAIT and there's still
 * no response, we'll try a ping, etc.
 *
 * If you're on a really slow net, this might need to be cranked up
 * depending on the idle timer interval, too. (ie, we'll wait a maximum
 * of twenty seconds for a response with these defaults.)
 */
#define SH_Q_MAX_WAIT       10 

extern int sh_q_initialized;
extern sh_q_item_t *queues[SH_MAX_CLIENTS];

/* initialize queue data structures; call this once upon startup */
int sh_q_init(void);

/* frees queue entries */
void sh_q_cleanup(void);

/* append an item to the queue for client n */
int sh_q_append(int, sh_q_item_t*);

/* prepend an item to the queue for client n */
int sh_q_prepend(int, sh_q_item_t*);

/* handle input for client n */
int sh_q_got_input(int client, int seq, sh_q_item_t **orig);

/* check pending queue items (idle loop) */
gint sh_q_idle(gpointer);

#endif 

